"""Test cases based on gNMI Subscribe ON requests."""

from oc_config_validate import schema, testbase


class SubsOnChangeTestCase(testbase.TestCase):
    """"""
    notifications_before_sync = []
    notifications_after_sync = []
    xpath = ""
    assert_all_paths_updated = False

    def assertNotifications(self):
        """Asserts that the received Updates have different values.

        Args:
          assert_all_paths_updated: If True, it asserts that all paths in
          the initial Update message had a change after.
        """
        self.assertTrue(bool(self.notifications_before_sync),
                        "No gNMI Subscribe response before sync_response")
        initial_updates = {}
        for n in self.notifications_before_sync:
            for u in n.update:
                path = schema.pathToString(u.path)
                self.assertTrue(
                    schema.isPathInRequestedPaths(path, [self.xpath]),
                    f"Unexpected update path {path} for subscription")
                initial_updates[schema.pathToString(
                    u.path)] = schema.typedValueToPython(u.val)

        self.assertTrue(bool(self.notifications_after_sync),
                        "No gNMI Subscribe response after sync_response")
        last_updates = initial_updates.copy()
        changed_paths = set()
        for n in self.notifications_after_sync:
            for u in n.update:
                path = schema.pathToString(u.path)
                value = schema.typedValueToPython(u.val)
                self.assertNotEqual(
                    last_updates[path], value,
                    f"Update for {path} has the same previous value {value}")
                last_updates[path] = value
                changed_paths.add(path)

        if self.assert_all_paths_updated:
            paths_not_updated = set(
                initial_updates.keys()).difference(changed_paths)
            self.assertEqual(
                len(paths_not_updated), 0,
                f"No Updates after sync_response for paths: "
                f"{paths_not_updated}")


class SubscribeAndSet(SubsOnChangeTestCase):
    """Subscribes ON_CHANGE and checks the response elements.

    1. Subscribes ON_CHANGE to the path.
    1. Collects initial Notifications.
    1. Issues a gNMI Set request.
    1. Collects Notifications until timeout.

    Does not support heartbeat.
    Does not support OnlyUpdates, the initial update is needed for comparison.

    Checks that the paths in the inital Updates are within the Request path.
    Checks that the Update values after the change are different from before
      the gNMI Set.

    Args:
        xpath: gNMI path to subscribe to.
        set_xpath: gNMI path to change.
        set_value: Value to change to.
        timeout_secs: Seconds to wait to receive an Update after Set.
        assert_all_paths_updated: If True, it asserts that all paths in the
          initial Update message had a change after. Defaults to False.
    """
    xpath = ""
    set_xpath = ""
    set_value = None
    timeout_secs = 10

    def testSubscribeOnChange(self):
        """"""

        def changeSet():
            self.gNMISetUpdate(self.set_xpath, self.set_value)

        self.assertArgs(["xpath", "set_xpath", "set_value"])
        self.assertXpath(self.xpath)
        self.assertXpath(self.set_xpath)

        (self.notifications_before_sync,
         self.notifications_after_sync) = self.gNMISubsStreamOnChange(
            self.xpath, self.timeout_secs, changeSet)

        self.assertNotifications()


class Subscribe(SubsOnChangeTestCase):
    """Subscribes ON_CHANGE and checks the response elements.

    1. Subscribes ON_CHANGE to the path.
    1. Collects initial Notifications.
    1. Collects further Notifications until timeout.

    Does not support heartbeat.
    Does not support OnlyUpdates, the initial update is needed for comparison.

    Checks that the paths in the inital Updates are within the Request path.
    Checks that the Update values are different from previous.

    Args:
        xpath: gNMI path to subscribe to.
        timeout_secs: Seconds to wait to receive an Update after Set.
        assert_all_paths_updated: If True, it asserts that all paths in
          the initial Update message had a change after. Defaults to False.
    """
    xpath = ""
    timeout_secs = 10

    def testSubscribeOnChange(self):
        """"""

        self.assertArgs(["xpath"])
        self.assertXpath(self.xpath)

        (self.notifications_before_sync,
         self.notifications_after_sync) = self.gNMISubsStreamOnChange(
            self.xpath, timeout=self.timeout_secs)

        self.assertNotifications()



"""Test cases based on gNMI Get requests."""

import json

from oc_config_validate import schema, testbase


class Get(testbase.TestCase):
    """Tests that gNMI Get of an xpath returns no error.

    All arguments are read from the Test YAML description.

    Args:
        xpath: gNMI path to read.
    """
    xpath = ""

    def test0200(self):
        """"""
        self.assertArgs(["xpath"])
        self.assertXpath(self.xpath)
        self.gNMIGetAssertGet(self.xpath)


class GetCompare(testbase.TestCase):
    """Compares that gNMI Get of an xpath returns the expected value.

    All arguments are read from the Test YAML description.

    Args:
        xpath: gNMI path to read.
        want: Expected value; can be numeric, string or JSON-IETF.
    """

    xpath = ""
    want = ""

    @testbase.retryAssertionError
    def testGetCompare(self):
        """"""
        self.assertArgs(["xpath", "want"])
        self.assertXpath(self.xpath)
        resp_val = self.gNMIGetAssertGet(self.xpath)
        got = schema.typedValueToPython(resp_val)
        self.assertEqual(type(got), type(self.want),
                         "Values of different types")
        if isinstance(self.want, dict):
            cmp, diff = schema.intersectCmp(self.want, got)
            self.assertTrue(cmp,  diff)
        else:
            self.assertEqual(self.want, got)


class GetJsonCheck(testbase.TestCase):
    """Tests that gNMI Get of an xpath returns a schema-valid JSON response.

    All arguments are read from the Test YAML description.

    Args:
        xpath: gNMI path to read.
        model: Python binding class to check the JSON reply against.
    """

    xpath = ""
    model = ""

    @testbase.retryAssertionError
    def testGetJsonCheck(self):
        """"""
        self.assertArgs(["xpath", "model"])
        self.assertXpath(self.xpath)
        self.assertModelXpath(self.model, self.xpath)
        resp_val = self.gNMIGetAssertJson(self.xpath)
        model = schema.ocContainerFromPath(self.model, self.xpath)
        self.assertJsonModel(resp_val, model,
                             "Get response JSON does not match model")


class GetJsonCheckCompare(testbase.TestCase):
    """Checks for schema validity and compares a gNMI Get response.

    All arguments are read from the Test YAML description.

    Args:
        xpath: gNMI path to read.
        model: Python binding class to check the JSON reply against.
        want_json: Expected JSON-IETF value.
    """

    xpath = ""
    model = ""
    want_json = None

    @testbase.retryAssertionError
    def testGetJsonCheckCompare(self):
        """"""
        self.assertArgs(["xpath", "want_json", "model"])
        self.assertXpath(self.xpath)
        self.assertModelXpath(self.model, self.xpath)
        self.assertIsInstance(self.want_json, dict,
                              "'want_json' is not a valid JSON object")
        resp_val = self.gNMIGetAssertJson(self.xpath)
        model = schema.ocContainerFromPath(self.model, self.xpath)
        self.assertJsonModel(resp_val, model,
                             "Get response does not match the model")
        got = json.loads(resp_val)
        cmp, diff = schema.intersectCmp(self.want_json, got)
        self.assertTrue(cmp, diff)
